;****************************************************************
; Programname: L-Meter.asm
;
; For the ATMega48 @ 20 MHz
; Converts frequency to coil-value and shows it on the LCD
; Internal timer is driven by external LC oscillator. 
;
; L = 1 / (2.pi.Fr)^2 C
;
; Measures 100nH - 100 mH
;
;****************************************************************

.include 	"m48def.inc"		;ATMEGA48

;****************************************************************
; constants
;****************************************************************

; Default K for C = 3.5 nH = 7237227 = $6E6E6B
; Real K is calculated during calibration
.EQU	Kdh =	$6E
.EQU	Kdm =	$6E
.EQU	Kdl = 	$6B

; dLd is default correction offset in nH
; real dL is calculated during calibration
.EQU	dLd =   10		;nH

;in-range boundaries of L binairy
.EQU	Lmin	= 100
.EQU	Lmax	= 999

;Calibration signature
.EQU	signature	= $5A

;LowBat level = 6V x 0.0991 x 256 / 1.1
.EQU	LowBat = 138
		
;****************************************************************
; Variables
;****************************************************************
.DSEG

F:		.byte	3					;measured F/10 (in 100 mS)
K:		.byte	3					;3 bytes for calibrated K
L:		.byte	2					;inductance 100-999
Lbcd: 	.byte	3					;L decimal
Decade:	.byte	1					;measuring decade 0-5
OVER:	.byte	1					;Flag,1 = OVF (L > 99.9 mH)
DFsq:	.byte	3					;(D x F)^2 storage
Valid:	.byte	1					;1 = valid
dL:		.byte	1					;correction offset of measured L (nH)

;****************************************************************
; Interrupt vectors
;****************************************************************
.CSEG

.ORG	0		rjmp	Reset
.ORG	$10		rjmp	Timer0OVF
.ORG	$0D		rjmp	Timer1OVF

;****************************************************************
; Tables
;****************************************************************



;****************************************************************
;
; L = K / Fr^2   with K = 1 / C.(2.pi)^2
; 
; L = K / (D x Fr^2 x 10)   x 10 because F measure period = 100 mS
; D determines range F = Fr x 10
; 
; L is in range if K / (10 x D x Fr)^2 = 100 - 999
;*****************************************************************

;       D[decade]                           L = 100-999 means
D:
.DW		6554		;10 x 0.01 / 2^16			10.0 - 99.9 mH
.DW		2072		;10 x 0.0031623 / 2^16		1.00 - 9.99 mH
.DW		 655		;10 x 0.001 / 2^16			100  -  999 uH
.DW		 207		;10 x 0.00031623 / 2^16		10.0 -  99.9 uH
.DW		  66		;10 x 0.0001 / 2^16			1.00 -  9.99 uH
.DW		  21		;10 x 0.000031623 / 2^16	100  -   999 nH
.DW		3074		;D[6] = SQR(Kc) / F(/10)    if calibrated with 22uH, 3.5nH
					;so K := (D[6] x F)^2       for calibration with 22uH


Prefix:
.Db		'm','m','u','u','u','n'	;milli,micro,nano - Henry

CommaPos:
.DB		3,2,0,3,2,0				;commaposition in display, 0 = no comma	


;message strings
OVERstr:
.DB		"OVER    "

NOVALUEstr:
.DB		"NO VALUE"

SpacesStr:
.DB		"        "

P22uHstr:
.DB		"Place L1=22.0 uH"

P220nHstr:
.DB		" Place L2=220 nH"

CalOKstr:
.DB		" Calibration OK "

L1str:
.DB		"L1 out of range "

L2str:
.DB		"L2 out of range "

LowBatstr:
.DB		"   LOW BATTERY  "


;***************************************************************



;**************************************************************
; Timer0 Overflow Interrupt
; Measures LC-oscillatorfreq. / 10 (during 100mS)
; 8 bit timer0 counts from T0 input (PD4, Pin6)
; Fmax = 8 MHz
; 
; F = Fh,Fm,Fl,   Fh,Fm = Z   Fl = Timer0 
;**************************************************************
Timer0OVF:
	adiw	Zl,1

	reti
;**************************************************************



;**************************************************************
; Timer1OVF
;
; Occurs 100mS after starting Timer1
; Porttime for Freq.counter
;**************************************************************
Timer1OVF:
	ldi		r24,1

	reti
;**************************************************************



	
;**************************************************************
; Delay100uS
;**************************************************************
delay100uS:
	ldi		r17,250
dlu:
	nop
	nop
	nop
	nop
	nop
	dec		r17
	brne	dlu

	ret
;**************************************************************



;**************************************************************
; Reset
;
; Initalisations
;**************************************************************
Reset:
	;init stackpointer
	ldi		r16,low(Ramend)
	out		SPL,r16
	ldi		r16,high(Ramend)
	out		SPH,r16

	;load default K(3)
	ldi		r16,Kdl
	sts		K,r16
	ldi		r16,Kdm
	sts		K+1,r16
	ldi		r16,Kdh
	sts		K+2,r16

	;load default dL(1)
	ldi		r16,dLd
	sts		dL,r16

	;short delay for Vdd rise
	ldi		r16,25
	rcall	Delay10mS

	;init Ports
	ldi		r16,0
	out		DDRD,r16			;T0/PD4 is input (oscillator)
	ldi		r16,0b00111111
	out		DDRB,r16			;PB0-PB5 output (LCD)
	ldi		r16,0				
	out		DDRC,r16			;PC all inputs
	ldi		r16,(1<<5)
	out		PORTC,r16			;pullup on PC5 (S1 to mass)

	;init the LDC (HITACHI standard)
	;PB0-PB3 is LCD D4-D7 (4bit mode)
	;PB4 = RS (1=data,0=instruction)
	;PB5 is Enable (pulse)
	;LCD R/W = write always
	ldi		r16,0b00000011		;Reset
	rcall	WriteLCD
	ldi		r16,1
	rcall	Delay10mS

	ldi		r16,0b00000011		;Reset
	rcall	WriteLCD
	ldi		r16,1
	rcall	Delay10mS
	
	ldi		r16,0b00000011		;Reset
	rcall	WriteLCD
	ldi		r16,1
	rcall	Delay10mS

	ldi		r16,0b00000010		;4 bit operation
	rcall	WriteLCD

	ldi		r16,0b00000010
	rcall	WriteLCD
	ldi		r16,0b00001000		;2 lines, 5x8 display
	rcall	writeLCD

	ldi		r16,0b00000000
	rcall	WriteLCD
	ldi		r16,0b00001000		;display off, cursor off, blink off
	rcall	WriteLCD

	ldi		r16,0b00000000
	rcall	WriteLCD
	ldi		r16,0b00001100		;display on,cursor off, blink off
	rcall	WriteLCD

	ldi		r16,0b00000000
	rcall	WriteLCD
	ldi		r16,0b00000110		;cursor postion moves to the right
	rcall	WriteLCD

	ldi		r16,0b00000000
	rcall	WriteLCD
	ldi		r16,0b00000010		;Set DDRam adress 0, (most left position)
	rcall	WriteLCD
	ldi		r16,1
	rcall	Delay10mS

	ldi		r16,0b00000000
	rcall	WriteLCD
	ldi		r16,0b00000001		;clear the Display
	rcall	WriteLCD
	ldi		r16,1
	rcall	Delay10mS

	;init Timer0 (halted)
	ldi		r16,0b00000000
	out		TCCR0A,r16			;normal operation
	ldi		r16,0b00000000
	out		TCCR0B,r16			;stopped
	cli
	ldi		r16,0b00000001
	sts		TIMSK0,r16			;T0OVF int enabled

	;init Timer1 (halted)
	ldi		r16,0b00000000		;normal operation, no PWM
	sts		TCCR1A,r16
	ldi		r16,0b00000000		;timer stopped (no clk)
	sts		TCCR1B,r16
	ldi		r16,0				;no output compare 
	sts		TCCR1C,r16
	ldi		r16,1
	sts		TIMSK1,r16			;Timer1 OVF interrupt enabled
		
	;init the ADC (battery voltage test)
	ldi		r16,0b11100000		;internal 1.1V ref, adjust result left, ADC0 is input
	sts		ADMUX,r16
	ldi		r16,0b00000111		;adc off, adc stopped, no auto trigger
	sts		ADCSRA,r16			;no interrupt, CLK = sysclk / 128 (Tc = 160uS)
	


	rjmp	main
;**************************************************************





;**************************************************************
; WriteLCD
; Writes 4 bits to the LCD
; Entry r16, b0-3 (D4-D7), b4 = RS (1= data), b5 = enable
;**************************************************************
WriteLCD:
	out		PortB,r16
	sbi		PortB,5			;enable hi
	rcall	delay100uS
	cbi		PortB,5			;enable lo
	rcall	delay100uS

	ret
;***************************************************************



;***************************************************************
; Display
; Displays a character (cursor shifts automatically)
;
; Writes in 4 bit mode
; MSB is first
;
; Entry: r16, the character
; 
;***************************************************************
Display:
	;write MSB
	mov		r18,r16
	swap	r16			
	andi	r16,$0F
	ori		r16,(1<<4)	;bit4 =  RS (1=data)
	rcall	WriteLCD

	;write LSB
	mov		r16,r18
	andi	r16,$0F
	ori		r16,(1<<4)	;RS = 1
	rcall	WriteLCD

	ldi		r16,1
	rcall	Delay10mS	

	ret
;***************************************************************




;***************************************************************
; Delay10mS
;
; Delays r16 x 10 mS
;***************************************************************
Delay10mS:
	ldi		r18,200
Dl1:
	ldi		r17,200
Dl0:
	dec		r17
	nop
	nop
	brne	Dl0

	dec		r18
	brne	Dl1

	dec		r16
	brne	Delay10mS

	ret
;***************************************************************




;***************************************************************
; CalcKdivDFsq; 
;
; L(3) = K(3) / DFsq(3)
;
; uses: r10,r11,r12 (result)
; uses: r13,r14,r15 (denominator) 
; uses  r2,r3,r4    (numerator)
;***************************************************************
CalcKdivDFsq:
	ldi		r25,0		;bitcounter

	;clear result
	clr		r10
	clr		r11
	clr		r12

	;copy DFsq(3) to denominator
	lds		r13,DFsq
	lds		r14,DFsq+1
	lds		r15,DFsq+2

	;test if denominator = 0
	ldi		r16,1
	sts		Valid,r16
	ldi		r16,0
	cp		r13,r16
	brne	nZ
	cp		r14,r16
	brne	nZ
	cp		r15,r16
	brne	nZ
	clr		r16			;denominator = 0
	sts		Valid,r16	;return with Valid = 0

	ret
nZ:
	;copy K(3) to numerator
	lds		r2,K
	lds		r3,K+1
	lds		r4,K+2

cmpKD:
	;compare numerator - denominator
	mov		r16,r2
	sub		r16,r13
	mov		r16,r3
	sbc		r16,r14
	mov		r16,r4
	sbc		r16,r15
	
	brcc	gtr

	;numerator is smaller, shift denominator 1 right and start
	dec		r25
	brmi	Res2Small		;abort if K < DFsq (L=0)

	clc
	ror		r15
	ror		r14
	ror		r13

	rjmp	startbit
	
gtr:	
	;num > denom: shift denominator 1 bit to left

	sbrc	r15,7
	rjmp	startbit		;if den(MSB) = 1 then start dividing now

	
	;shift left denominator (r15,r14,r13)
	clc
	rol		r13
	rol		r14
	rol		r15

	inc		r25				;count shifts

	rjmp	cmpKD	;next compare K-D

startbit:
	;start deviding
	;numerator := numerator - shifted denominator (r4,r3,r2 - r15,r14,r13)

	sub		r2,r13
	sbc		r3,r14
	sbc		r4,r15 

	;set bit in result
	ldi		r16,1
bit0:
	or		r10,r16	

;	cpi		r25,0
;	breq	LRdy

	
	;shift denominator 1 right
	clc
	ror		r15
	ror		r14
	ror		r13

	cpi		r25,0
	breq	LRdy

	;ready if r25 = 0
	dec		r25

	;shift result 1 left
	clc		
	rol		r10
	rol		r11
	rol		r12

	;compare numerator- shifted denominator
	mov		r16,r2
	sub		r16,r13
	mov		r16,r3
	sbc		r16,r14
	mov		r16,r4
	sbc		r16,r15


	brcc	startbit		;resultbit = 1
	clr		r16
	rjmp	bit0			;resultbit = 0
	

	;L < 1
Res2Small:
	clr		r16
	sts		L,r16
	sts		L+1,r16
	
	ret

LRdy:
	;round the result

	;compare numerator- shifted denominator
	mov		r16,r2
	sub		r16,r13
	mov		r16,r3
	sbc		r16,r14
	mov		r16,r4
	sbc		r16,r15
	brcs	LRdy1
	
	;round up
	ldi		r16,1
	add		r10,r16
	ldi		r16,0
	adc		r11,r16

LRdy1:
	;copy result to L
	sts		L,r10
	sts		L+1,r11

	ret

;***************************************************************





;**************************************************************
; CalcK
;
; Calculates the calibrated K, for L = 22.0 uH
; Entry: Zh,Zl,Timer0 = Freq(/10) H,M,L
; Exit K(3)
;
; -> K(22u) = (D[6] x F)^2
;**************************************************************
CalcK:
	;measure frequency with 22.0 uH
	rcall	measureF
	;calculate DFsq with decade = 6
	ldi		r16,6
	sts		decade,r16
	rcall	calcDFsq
	;store calibration result in K
	lds		r16,DFsq
	sts		K,r16
	lds		r16,DFsq+1
	sts		K+1,r16
	lds		r16,DFsq+2
	sts		K+2,r16

	ret
;**************************************************************




;**************************************************************
; Tested OK
;
; CalcDFsq
;
; Calculates SQR(D[decade] x F(3))
; Entry: Decade 0-5
; Entry: D[decade] x 2^16
; Entry: F(3)
;
; Exit DFsq(3)
; Exit Valid 1/0
;**************************************************************
CalcDFsq:
	ldi		r16,1
	sts		Valid,r16

	;calculate D x F

	;clear result
	clr		r10
	clr		r11
	clr		r12
	clr		r13
	clr		r14

	;load Dl[decade*2]
	ldi		Zl,low(D*2)
	ldi		Zh,high(D*2)
	lds		r16,decade
	lsl		r16				;tablewidth is 2 bytes
	add		Zl,r16
	brcc	C00
	inc		Zh
C00:
	lpm		r16,Z+		;Dl

	;Dl x F(3) 
	lds		r18,F
	mul		r18,r16
	mov		r10,r0
	mov		r11,r1

	lds		r18,F+1
	mul		r18,r16
	add		r11,r0
	adc		r12,r1

	lds		r18,F+2
	mul		r18,r16
	add		r12,r0
	adc		r13,r1

	;load Dh[Decade*2]
	lpm		r16,Z		;Dh

	;Dh x F(3)
	lds		r18,F
	mul		r18,r16
	add		r11,r0
	adc		r12,r1

	lds		r18,F+1
	mul		r18,r16
	add		r12,r0
	adc		r13,r1

	lds		r18,F+2
	mul		r18,r16
	add		r13,r0
	adc		r14,r1

	;here D x F  = r14,r13,r12 (see table, D was D x 2^16)
	
	;round D x F
	sbrs	r11,7
	rjmp	lab0
	ldi		r16,1
	add		r12,r16
	clr		r16
	adc		r13,r16
	adc		r14,r16
	
lab0:
	;calculate (D x F)^2
	;result is 5 bytes, if not(b4=0 and b3=0) then DFsqValid = 0 (wrong decade) 

	;clr result
	clr		r3
	clr		r4
	clr		r5		
	clr		r6		
	clr		r7		

	;   r14,r13,r12
	;   r14,r13,r12
	;-------------- x
	;r7,r6,r5,r4,r3

	mul		r12,r12
	mov		r3,r0
	mov		r4,r1

	mul		r12,r13
	add		r4,r0
	adc 	r5,r1

	mul		r12,r14
	add		r5,r0
	adc		r6,r1

	mul		r13,r12
	add		r4,r0
	adc		r5,r1

	mul		r13,r13
	add		r5,r0
	adc		r6,r1

	mul		r13,r14
	add		r6,r0
	adc		r7,r1

	mul		r14,r12
	add		r5,r0
	adc		r6,r1

	mul		r14,r13
	add		r6,r0
	adc		r7,r1

	mul		r14,r14
	add		r7,r0
	
	;test if valid
	ldi		r16,0
	cp		r16,r1
	brne	invalid
	cp		r16,r7
	brne	invalid
	cp		r16,r6
	brne	invalid

	;store if valid
	sts		DFsq,r3
	sts		DFsq+1,r4
	sts		DFsq+2,r5

	ret

invalid:
	ldi		r16,0
	sts		Valid,r16			;
	
	ret
	
;**************************************************************



;**************************************************************
; CalcL
; calculates the inductance
;
; entry: Zh,Zl,Timer0  (Timer0 has stopped after 100mS)
; entry: K(3)
; entry: D[0..5]:Q15, possible decades (1 to 6)
; exit: L(2) binary
; exit: decade = 0 - 5 (determines prefix and commaposition)
;
; L = K / (D[decade] x F)^2, decade is valid if L is in range 100-999
;***************************************************************
CalcL:
	clr		r19		;start with decade 0
	
	clr		r16
	sts		OVER,r16

clcL:
	;denominator := (D x F)^2
	sts		decade,r19
	rcall	CalcDFsq
	lds		r16,Valid
	cpi		r16,1
	brne	NextDecade		;denominator > $FFFFFF means L < 1 (wrong decade)

	;calculate L = K / (D x F)^2
	rcall	CalcKdivDFsq		;result = L(2)
	lds		r16,Valid
	cpi		r16,1				;Valid = 0 if division by zero (F=0, no oscillation)
	brne	NoOsc

	;if (L < 100) then
	;  if (decade < 5) then nextdecade
	;  else L=result //allow below 100nH
	ldi		r20,low(Lmin)
	ldi		r21,high(Lmin)
	clr		r22

	;Test L-Lmin
	lds		r16,L
	sub		r16,r20
	lds		r16,L+1
	sbc		r16,r21

	brcs	nextdecade			;result was < 100

 	rjmp	Lfound				;first decade found with L > 100


	;next decade
NextDecade:
	cpi		r19,5
	breq	Lfound

	inc		r19
	rjmp	clcL


	;100 <= L(2) <= 999
Lfound:
	;if L > 999 and decade = 0 then OVER := 1
	cpi		r19,0
	brne	Lf1

	ldi		r20,low(Lmax)
	ldi		r21,high(Lmax)
	clr		r22

	;compare L - Lmax
	lds		r16,L
	sub		r16,r20
	lds		r16,L+1
	sbc		r16,r21

	brcs	Lf1			;L < Lmax

	ldi		r16,1
	sts		OVER,r16	

Lf1:
	;compensate with calibrated dL if decade = 5 (100-999nH range)
	lds		r16,decade
	cpi		r16,5
	brne	Lf2
	;decade is 5, L := L - dL
	lds		r16,dL
	lds		r17,L
	sub		r17,r16
	sts		L,r17			

	lds		r17,L+1
	clr		r16
	sbc		r17,r16
	sts		L+1,r17

Lf2:
	ret

NoOsc:	
	ldi		r16,2
	sts		OVER,r16

	ret
;***************************************************************




;***************************************************************
; MeasureF
;
; Measures LC frequency x 10 Hz 
; exit: F(3)
;***************************************************************
MeasureF:
	;clr Z and Timer0
	clr		Zh
	clr		Zl
	ldi		r16,0
	out		TCNT0,r16

	;prepare timer1 (100mS OVF)
	clr		r24						;100mS OVF flag

	ldi		r16,1
	out		GTCCR,r16				;timer1 prescaler reset

	;load timer1 value
	ldi		r16,low($10000-31250)	;load Upcounter ($10000 = OVF value)
	sts		TCNT1L,r16				;31250 x 64 / 20E6 = 100 mS
	ldi		r16,high($10000-31250)
	sts		TCNT1H,r16					

	ldi		r16,0b00000011
	sts		TCCR1B,r16			;Timer1 start counting, CLK = CLKio/64
	
	;Prepare timer0 (Freq counter)
	ldi		r16,0b00000111		;clk source = T0 pin, rising edge
	out		TCCR0B,r16

	sei							;start Freq.counter and 100mS timer

	;wait for Timer1 interrupt (100mS)
wt100mS:
	sbrs	r24,0
	rjmp	wt100mS

	;stop timer0
	ldi		r16,0b00000000		;no clk source (halted)
	out		TCCR0B,r16
	cli							

	;stop timer1
	ldi		r16,0b00000000
	sts		TCCR1B,r16			;Timer1 start counting, CLK = CLKio/64

	;store result in F(3)
	in		r16,TCNT0
	sts		F+0,r16		;L
	sts		F+1,Zl		;M
	sts		F+2,Zh		;H

	ret

;***************************************************************


;***************************************************************
; L2Lbcd
;
; Converts binairy L(2) to decimal Lbcd(3)
;***************************************************************
digit:
	.db		100,10,1,$0
L2Lbcd:
	clr		r16
	sts		Lbcd,r16
	sts		Lbcd+1,r16
	sts		Lbcd+2,r16

	ldi		Zl,low(digit*2)
	ldi		Zh,high(digit*2)

	lpm		r0,Z+

	ldi		Yl,low(Lbcd+2+1)		;+1 for pre-dec
	ldi		Yh,high(Lbcd+2+1)		;MSD first

	clr		r23						;digit value

	clr		r22						;digit counter (0,1,2)

trydigit:	
	;compare L with digit (100's, 10's, 1's)
	lds		r16,L
	sub		r16,r0
	lds		r16,L+1
	sbci	r16,0
	
	brcs	nextdigit

	inc		r23
	
	;L - digitvalue
	lds		r16,L
	sub		r16,r0
	sts		L,r16
	lds		r16,L+1
	sbci	r16,0
	sts		L+1,r16
	
	rjmp	trydigit		

nextdigit:
	st		-Y,r23		;store digitvalue in Lbcd
	lpm		r0,Z+
	clr		r23
	inc		r22
	cpi		r22,3		;last digit?
	brne	trydigit
	
	;correct if truncation-error
	;(due to limited accuracy of tablevalues)
	lds		r16,Lbcd+2
	cpi		r16,9+1
	brcs	noerr
	ldi		r16,9		
	sts		Lbcd,r16
	sts		Lbcd+1,r16
	sts		Lbcd+2,r16
	
noerr:

	ret
;***************************************************************



	
;***************************************************************
; Line1
;
; Puts cursor at pos 0 in line1
;***************************************************************
Line1:
	ldi		r16,0b00000000
	rcall	writeLCD
	ldi		r16,0b00000010
	rcall	writeLCD			;cursor 0, line1
	ldi		r16,1
	rcall	delay10mS

	ret
;***************************************************************




;***************************************************************
; Line2
;
; Puts cursor in LCD to begin of Line2 (offset $40)
;***************************************************************
Line2:
	;set the instruction
	ldi		r17,0b10000000		;set DDRam adress instruction
	ori		r17,$40				;set DDRam adress to pos 0 in line2

	;write the instruction
	mov		r16,r17
	swap	r16
	andi	r16,$0f
	rcall	WriteLCD			;write MSB

	mov		r16,r17
	andi	r16,$0f
	rcall	WriteLCD			;write LSB

	ldi		r16,1
	rcall	Delay10mS

	ret

;***************************************************************



;***************************************************************
; ClearLine
;
; Writes 8 spaces to selected line
; (gives smoother display then Cleardisplay command)
;***************************************************************
ClearLine:
	ldi		Zl,low(SpacesStr*2)
	ldi		Zh,high(SpacesStr*2)
	rcall	DisplayMessage

	ret
;***************************************************************



;***************************************************************
; DisplayLbcd
;
; Displays Inductance with comma and prefix
; Entry: Lbcd (3) , Lbcd+2 = MSD
; Entry: Decade = 0-5
;***************************************************************
DisplayLbcd:

	rcall	line2			;display L in right half of LCD
	
	;get commaposition
	ldi		Zl,low(Commapos*2)
	ldi		Zh,high(Commapos*2)

	lds		r16,Decade
	add		Zl,r16
	brcc	Di00
	inc		Zh
Di00:
	lpm				
	mov		r20,r0		;r20 is commaposition

	;display Lbcd (and comma)
	lds		r16,Lbcd+2
	subi	r16,-'0'
	cpi		r16,'0'
	brne	Di001
	ldi		r16,' '		;write space to leading zero (for L < 100 nH)
Di001:
	rcall	Display
	cpi		r20,2
	brne	Di01
	ldi		r16,'.'
	rcall	Display

Di01:
	lds		r16,Lbcd+1
	subi	r16,-'0'
	rcall	Display
	cpi		r20,3
	brne	Di02
	ldi		r16,'.'
	rcall	Display

Di02:
	lds		r16,Lbcd
	subi	r16,-'0'
	rcall	Display

	;space
	ldi		r16,' '
	rcall	Display

	cpi		r20,0
	brne	Di04
	
	;extra space if no comma
	ldi		r16,' '
	rcall	Display

Di04:
	;display the prefix and 'H'
	ldi		Zl,low(prefix*2)
	ldi		Zh,high(prefix*2)
	lds		r16,Decade
	add		Zl,r16
	brcc	Di03
	inc		Zh
Di03:
	lpm
	mov		r16,r0
	rcall	Display

	ldi		r16,'H'
	rcall	Display

	ret

;***************************************************************



;***************************************************************
; Displaymessage
;
; Displays a 8 character long message
; Line must be selected by Line1 or Line2 routine call
;
; Entry Zh,Zl: pointer to message adress
;***************************************************************
DisplayMessage:
	ldi		r26,8
DM00:
	lpm		r16,Z+
	rcall	Display
	dec		r26
	brne	DM00

	ret

;***************************************************************
		



;***************************************************************
; S1re
;
; returns if S1 released
; S1 = PC5 to mass
;***************************************************************
S1re:
	sbis	PinC,5
	rjmp	S1re
	ldi		r16,1
	rcall	Delay10mS	;debounce
	sbis	PinC,5
	rjmp	S1re

	ret
;***************************************************************




;***************************************************************
; S1pr
;
; returns if S1 pressed
; S1 = PC5 to mass
;***************************************************************
S1pr:
	sbic	PinC,5
	rjmp	S1pr
	ldi		r16,1
	rcall	Delay10mS	;debounce
	sbic	PinC,5
	rjmp	S1pr

	ret
;***************************************************************



;***************************************************************
; WriteEEPROM
;
; EEPROM[Zl] := 16
; Zl =  addr 0 to 255
;***************************************************************
WriteEEPROM:

EEPROM_write:
	; Wait for completion of previous write
	sbic 	EECR,EEPE
	rjmp 	EEPROM_write
	; Set up address in address register
	clr		r17
	out 	EEARH,r17
	out 	EEARL,Zl
	; Write data (r16) to Data Register
	out 	EEDR,r16
	; Write logical one to EEMPE
	sbi 	EECR,EEMPE
	; Start eeprom write by setting EEPE
	sbi 	EECR,EEPE
	
	ret

;***************************************************************



;***************************************************************
; ReadEEPROM
;
; r16 := EEPROM[Zl] 
; Zl = addr 0 to 255
;***************************************************************
ReadEEPROM:

EEPROM_read:
	; Wait for completion of previous write
	sbic	EECR,EEPE
	rjmp 	EEPROM_read
	; Set up address in address register
	clr		r17
	out 	EEARH,r17
	out 	EEARL,Zl
	; Start eeprom read by writing EERE
	sbi 	EECR,EERE
	; Read data from Data Register
	in 		r16,EEDR
	
	ret

;***************************************************************




;***************************************************************
; WriteCal2EEPROM
;
; Write Signature(1),K(3) and dL(1) to EEPROM[0-4]
;***************************************************************
WriteCal2EEPROM:

	;write signature
	ldi		r16,signature
	ldi		Zl,0			;start with adres 0
	rcall	WriteEEPROM

	;write K(3)
	lds		r16,K
	ldi		Zl,1
	rcall	WriteEEPROM

	lds		r16,K+1
	ldi		Zl,2
	rcall	WriteEEPROM

	lds		r16,K+2
	ldi		Zl,3
	rcall	WriteEEPROM

	;write dL(1)
	lds		r16,dL
	ldi		Zl,4
	rcall	WriteEEPROM

	ret	

;***************************************************************




;***************************************************************
; Calibrate
;
; Measure the calibration values K and dL
; and store them in EEPROM. Also write calibration signature
;
; Activated after startup with S1 pressed
;***************************************************************
Calibrate:
	;display 'Place L1=22.0 uH'
	rcall	line1
	ldi		Zl,low(P22uHstr*2)
	ldi		Zh,high(P22uHstr*2)
	rcall	DisplayMessage
	rcall	line2
	rcall	DisplayMessage

	;wait for S1 released
	rcall	S1re

	;wait for S1 pressed
	rcall	S1pr

	;measure F
	rcall	MeasureF
	;calculate L1
	rcall	CalcL
	;test if L valid
	lds		r16,over
	cpi		r16,0
	brne	L1err

	;Test if Decade = 3 (10.0 uH - 99.9 uH)
	lds		r16,Decade
	cpi		r16,3
	brne	L1err
	
	;Calculate K
	rcall	CalcK


	;display ' Place L2=220 nH'
	rcall	line1
	ldi		Zl,low(P220nHstr*2)
	ldi		Zh,high(P220nHstr*2)
	rcall	DisplayMessage
	rcall	line2
	rcall	DisplayMessage

	;wait for S1 released
	rcall	S1re

	;Wait for S1 pressed
	rcall	S1pr

	;measure L2
	rcall	MeasureF
	ldi		r16,0
	sts		dL,r16			;calculate L with dL = 0
	rcall	CalcL
	lds		r16,over
	cpi		r16,0
	brne	L2err

	;test if L2 > 220nH
	lds		r16,decade
	cpi		r16,5
	brne	L2err

	lds		r16,L+1
	brne	L2err
	lds		r16,L
	cpi		r16,220
	brmi	L2err

	;L2 is valid, Calculate dL = L2 - 220 (nH)
	subi	r16,220
	sts		dL,r16

	;Write signature(1), K(3) and dL(1) to EEPROM
	rcall	WriteCal2EEPROM

	;Display 'Calibrated OK' for 2.5 seconds
	rcall	line1
	ldi		Zl,low(CalOKstr*2)
	ldi		Zh,high(CalOKstr*2)
	rcall	DisplayMessage
	rcall	line2
	rcall	DisplayMessage

	ldi		r16,250
	rcall	Delay10mS

	ret

L2err:
	;L2 is invalid
	;Display 'L2 out of range' for 2.5 seconds
	rcall	line1
	ldi		Zl,low(L2str*2)
	ldi		Zh,high(L2str*2)
	rcall	DisplayMessage
	rcall	line2
	rcall	DisplayMessage

	ldi		r16,250
	rcall	Delay10mS

	ret

L1err:
	;L1 is invalid
	;Display 'L1 out of range' for 2.5 seconds
	rcall	line1
	ldi		Zl,low(L1str*2)
	ldi		Zh,high(L1str*2)
	rcall	DisplayMessage
	rcall	line2
	rcall	DisplayMessage

	ldi		r16,250
	rcall	Delay10mS

	ret
	
;***************************************************************




;***************************************************************
; GetValues
;
; If calibration signature present in EEPROM 
; then load K(3) and dL(1) from EEPROM[1-4]
; Else keep defaults set in Reset routine
;***************************************************************
GetValues:
	clr		Zl				;addres 0
	rcall	ReadEEPROM
	cpi		r16,Signature
	brne	Defaults

	;read K(3)
	ldi		Zl,1
	rcall	ReadEEPROM
	sts		K,r16

	ldi		Zl,2
	rcall	ReadEEPROM
	sts		K+1,r16

	ldi		Zl,3
	rcall	ReadEEPROM
	sts		K+2,r16

	;read dL(1)
	ldi		Zl,4
	rcall	ReadEEPROM
	sts		dL,r16

Defaults:
	;keep defaults (set in RESET routine)
	;(run uncalibrated)
	ret
;***************************************************************




;***************************************************************
; TestBattery
;
; Measures the batteryvoltage
; Displays Low Battery message if voltage too low
;***************************************************************
TestBattery:
	;start ADC
	ldi		r16,0b11000111		;adc on, conversion start
	sts		ADCSRA,r16			;no interrupt, CLK = sysclk / 128 (Tc = 160uS)

	;wait for conversion complete
W4ADC:
	lds		r16,ADCSRA
	sbrc	r16,6
	rjmp	W4ADC

	;compare with low level
	lds		r16,ADCH
	cpi		r16,LowBat
	brpl	TB00

	;show message if level too low
	rcall	line1
	ldi		Zl,low(LowBatStr*2)
	ldi		Zh,high(LowBatStr*2)
	rcall	DisplayMessage
	rcall	line2
	rcall	DisplayMessage

	ldi		r16,100
	rcall	Delay10mS
TB00:
	ret
;***************************************************************




;***************************************************************
; Main
;
; Mainflow. Calculate and display the Inductance
;***************************************************************
Main:
	rcall	MeasureF	;initial

	clr		r27		;batterytest event counter

	;if S1 pressed then Calibrate
	sbic	PinC,5
	rjmp	Mn00
	rcall	Calibrate
	rcall	line2
	rcall	ClearLine
	rjmp	Measure
Mn00:
	;If calibrate signature present in EEPROM load K and dF
	;Else load default K and dF
	rcall	getValues

Measure:
	;Test battery each 10 seconds
	inc		r27
	cpi		r27,35
	brne	Msr00
	clr		r27
	rcall	TestBattery

Msr00:
	;measure the frequency
	rcall	MeasureF

	;calculate the Inductance
	rcall	CalcL

	;test if OVER=1 (L > 99.9 mH)
	lds		r16,over
	cpi		r16,1
	breq	ovr

	;test if OVER=2 (No oscillation)
	cpi		r16,2
	breq	NoValue

	;L value is valid
	;overwrite errormessage (clear instruction causes flickering LDC)
	rcall	 line1

	;clear line1 with spaced
	;(clear instr. causes flickering LCD)
	ldi		r26,8
cld1:
	ldi		r16,' '
	rcall	display
	dec		r26
	brne	cld1

	;convert to decimal
	rcall	L2Lbcd

	;show the Inductance on LCD
	rcall	DisplayLbcd

	rjmp	measure				;measure continuous

ovr:
	rcall	line2
	rcall	ClearLine

	;OVER = 1, display 'OVER'
	rcall	Line1				;line1, cursor 0
	;write 'OVER    '
	ldi		Zl,low(OVERstr*2)
	ldi		Zh,high(OVERstr*2)
	rcall	DisplayMessage
	
	rjmp	measure

NoValue:
	;clear line2 (clear instr. causes flickering LCD)
	rcall	Line2
	rcall	ClearLine

	;OVER = 2, display 'NO VALUE' (no oscillation possible)
	rcall	line1				;line1, cursor 0

	;write 'NO VALUE'
	ldi		Zl,low(NOVALUEstr*2)
	ldi		Zh,high(NOVALUEstr*2)
	rcall	DisplayMessage

	rjmp	measure
			
;***************************************************************


